(***********************************************************************)
(*                                                                     *)
(*                               Grimage                               *)
(*                                                                     *)
(*                           Graphics image                            *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2000-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: colors.ml,v 1.17 2017/05/31 14:42:47 fclement Exp $ *)

(* Definitions and manipulations of colors. *)

type t = Graphics.color;;

type rgb_channel = int;;

type red = rgb_channel
and green = rgb_channel
and blue = rgb_channel
;;

type hsb_channel = int;;

type hue = hsb_channel
and saturation = hsb_channel
and brightness = hsb_channel
;;

(* Extra graphics functions primitives. *)
let nround x y = (2 * x + y) / (2 * y);;
let mxround x = nround (0xFF * x) 0xFF;;

(* color -> rgb. *)
let red_of_color c = (c lsr 16) land 0xFF;;
let green_of_color c = (c lsr 8) land 0xFF;;
let blue_of_color c = c land 0xFF;;

let r_of = red_of_color;;
let g_of = green_of_color;;
let b_of = blue_of_color;;

let rgb_of_color c = (r_of c, g_of c, b_of c);;

(* rgb -> color. *)
let color_of_rgb r g b = (r lsl 16) + (g lsl 8) + b;;
let rgb = color_of_rgb;;

(* rgb -> hsb. *)
let brightness_of_rgb r g b = max r (max g b);;

let get_min_and_dyn bri r g b =
  let min = min r (min g b) in
  (min, bri - min)
;;

let get_sat bri r g b =
  let (_, d) = get_min_and_dyn bri r g b in
  if bri = 0 then 0 else nround (0xFF * d) bri
;;

let saturation_of_rgb r g b = get_sat (brightness_of_rgb r g b) r g b;;

let get_hue bri r g b =
  let (min, d) = get_min_and_dyn bri r g b in
  if bri = 0 || d = 0 then 0 else
  let r' = (bri - r) * 0xFF / d
  and g' = (bri - g) * 0xFF / d
  and b' = (bri - b) * 0xFF / d in
  let h =
    if bri = r then if min = g then 5 * 0xFF + b' else 0xFF - g' else
    if bri = g then if min = b then 0xFF + r' else 3 * 0xFF - b' else
    if min = r then 3 * 0xFF + g' else 5 * 0xFF - r' in
  h / 6
;;

let hue_of_rgb r g b = get_hue (brightness_of_rgb r g b) r g b;;

let hsb_of_rgb r g b =
  let bri = brightness_of_rgb r g b in
  let sat = get_sat bri r g b in
  let hue = get_hue bri r g b in
  (hue, sat, bri)
;;

(* hsb -> rgb. *)
let rgb_of_hsb h s b =
  let h = h * 6 in
  let i = h / 0xFF * 0xFF in
  let f = h - i in
  let m = b * (0xFF - s) / 0xFF and n = b * (0xFF - s * f / 0xFF) / 0xFF
  and k = b * (0xFF - s * (0xFF - f) / 0xFF) / 0xFF in
  match i / 0xFF with
  | 0 | 6 -> (mxround b, mxround k, mxround m)
  | 1 -> (mxround n, mxround b, mxround m)
  | 2 -> (mxround m, mxround b, mxround k)
  | 3 -> (mxround m, mxround n, mxround b)
  | 4 -> (mxround k, mxround m, mxround b)
  | 5 -> (mxround b, mxround m, mxround n)
  | _ -> invalid_arg "rgb_of_hsb"
;;

(* hsb -> color. *)
let color_of_hsb h s b =
  let (r, g, b) = rgb_of_hsb h s b in
  color_of_rgb r g b
;;

let hsb = color_of_hsb;;

(* color -> hsb. *)
let hue_of_color c =
  let (r, g, b) = rgb_of_color c in
  hue_of_rgb r g b
;;

let saturation_of_color c =
  let (r, g, b) = rgb_of_color c in
  saturation_of_rgb r g b
;;

let brightness_of_color c =
  let (r, g, b) = rgb_of_color c in
  brightness_of_rgb r g b
;;

let hsb_of_color c =
  let (r, g, b) = rgb_of_color c in
  hsb_of_rgb r g b
;;

let hue_of = hue_of_color;;
let sat_of = saturation_of_color;;
let bri_of = brightness_of_color;;

(* Color transformations. *)

let hue c =
  let hue = hue_of c in
  color_of_hsb hue 0xFF 0xFF
;;

type shift = int;;

let shift_hue_of_hsb s hue = (hue + s) mod 0xFF;;

let shift_hue s c =
  let (hue, sat, bri) = hsb_of_color c in
  let hue' = shift_hue_of_hsb s hue in
  color_of_hsb hue' sat bri
;;

type gray = int;;

let rgb_of_gray gr = (gr, gr, gr);;
let color_of_gray gr = color_of_rgb gr gr gr;;

let lift_gray f c =
  let gr = f c in
  color_of_gray gr
;;

let red_gray = lift_gray r_of
and green_gray = lift_gray g_of
and blue_gray = lift_gray b_of
;;

let lift_gray_from_rgb f_rgb c =
  let (r, g, b) = rgb_of_color c in
  let gr = f_rgb r g b in
  color_of_gray gr
;;

let hue_gray_of_rgb = hue_of_rgb;;
let hue_gray = lift_gray_from_rgb hue_gray_of_rgb;;

let saturation_gray_of_rgb = saturation_of_rgb;;
let saturation_gray = lift_gray_from_rgb saturation_gray_of_rgb;;

let brightness_gray_of_rgb = brightness_of_rgb;;
let brightness_gray  = lift_gray_from_rgb brightness_gray_of_rgb;;

let intensity_gray_of_rgb r g b = (r + g + b) / 3;;
let intensity_gray = lift_gray_from_rgb intensity_gray_of_rgb;;

let lightness_gray_of_rgb r g b =
  let bri = brightness_of_rgb r g b in
  let (min, _) = get_min_and_dyn bri r g b in
  (bri + min) / 2
;;
let lightness_gray = lift_gray_from_rgb lightness_gray_of_rgb;;

let luma_gray_of_rgb r g b = (299 * r + 587 * g + 114 * b) / 1000;;
let luma_gray = lift_gray_from_rgb luma_gray_of_rgb;;

type hue_range = hue * hue;;

type in_hue_range = hue_range
and out_hue_range = hue_range
;;

let is_empty_hue_range (hue1, hue2) = hue2 < hue1;;

let valid_hue_range (hue1, hue2) = hue2 - hue1 <= 0xFF;;

let is_in_integer_range (hue1, hue2) hue = hue1 <= hue && hue < hue2;;

let is_not_in_hue_range rg hue =
  not (is_in_integer_range rg hue) &&
  not (is_in_integer_range rg (hue + 0xFF))
;;

let linear_integer_map ((hue1, hue2), (hue1', hue2')) hue =
  (hue - hue1) * (hue2' - hue1') / (hue2 - hue1) + hue1'
;;

let linear_hue_map (in_rg, _ as rgs) hue =
  let hue' =
    if is_in_integer_range in_rg hue then linear_integer_map rgs hue else
    let hue = hue + 0xFF in
    if is_in_integer_range in_rg hue then linear_integer_map rgs hue else
    assert false in
  hue' mod 0xFF
;;

let gen_select to_hsb of_hsb (in_rg, out_rg as rgs) =
  if is_empty_hue_range in_rg then invalid_arg "empty input hue range" else
  if is_empty_hue_range out_rg then invalid_arg "empty output hue range" else
  if not (valid_hue_range in_rg) then
    invalid_arg "too large input hue range" else
  if not (valid_hue_range out_rg) then
    invalid_arg "too large output hue range" else
  (fun gmap x ->
    let (hue, sat, bri) = to_hsb x in
    if is_not_in_hue_range in_rg hue then gmap x else
    let hue' = linear_hue_map rgs hue in
    of_hsb hue' sat bri)
;;

(* let curry f r g b = f (r, g, b);; *)
let uncurry f (r, g, b) = f r g b;;

let select_color_of_rgb rgs gmap_of_rgb r g b =
  let gmap_of_rgb rgb = rgb_of_gray ((uncurry gmap_of_rgb) rgb) in
  gen_select (uncurry hsb_of_rgb) rgb_of_hsb rgs gmap_of_rgb (r, g, b)
;;

let select_color rgs gmap c = gen_select hsb_of_color color_of_hsb rgs gmap c;;
