(***********************************************************************)
(*                                                                     *)
(*                              Ref-image                              *)
(*                                                                     *)
(*                  Image Segmentation by Refinement                   *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2009-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: configuration.ml,v 1.54 2017/05/31 16:38:14 fclement Exp $ *)

module Nat = Unatural;;

module Opt = Options;;

let c_qual = Upervasives.qualify "Configuration";;

let c_invalid_arg = Upervasives.invalid_arg_qual c_qual "";;

open Segmentation_types;;
open Configuration_types;;
open Configuration_types_print;;

(*let gen_digest = Utype_digest.gen_digest Configuration_digest.md5;;
let digest = gen_digest "configuration";;*)

(* History functor interface. *)

let printer = printer_t;;
let print = print_t;;

let get_dimension () = Nat.of_int (Opt.get_dimension ());;

let get_dimension_management () =
  match Opt.get_dimension_management () with
  | Opt.Vector -> (Best_cutting, All_components)
  | Opt.Best_component_only -> (Best_cutting, Best_component)
  | Opt.Best_component_for_each -> (All_cuttings, Best_component)
  | Opt.Combine_best_components -> (All_cuttings, All_components)
;;

let get_cutting_strategy () =
  match Opt.get_cutting_strategy () with
  | Opt.Option_Best_in_a_family ->
      let nct = Nat.of_int (Opt.get_number_of_cuts ()) in
      let bcfl = Opt.get_basic_cutting_families () in
      Best_in_a_family (nct, Ulist1.of_list bcfl)
  | Opt.Option_Overall_best -> Overall_best
;;

let get_stopping_criterion () =
  {
   max_iteration = Nat.of_int (Opt.get_max_iteration ());
   min_best_scalar_indicator = Opt.get_min_best_scalar_indicator ();
   min_scalar_relative_error = Opt.get_min_scalar_relative_error ();
  }
;;

let get_image_width = Opt.get_argument_initial_image_width;;
let get_image_height = Opt.get_argument_initial_image_height;;

let { Uaref.get = get_configuration_opt; set = set_configuration_opt; } =
  Uaref.make_aref_unique "algorithm configuration" None
;;

let cold_set () =
  let config =
    {
     dimension = get_dimension ();
     dimension_management = get_dimension_management ();
     cutting_strategy = get_cutting_strategy ();
     stopping_criterion = get_stopping_criterion ();
     image_width = get_image_width ();
     image_height = get_image_height ();
   } in
  set_configuration_opt (Some config)
;;

let warm_set config =
  let config =
    { config with stopping_criterion = get_stopping_criterion (); } in
  set_configuration_opt (Some config)
;;

let get () =
  match get_configuration_opt () with
  | None -> assert false
  | Some config -> config
;;

(* Algorithm configuration. *)

let phase_dimension = Options.phase_dimension;;

let get_dimension () = (get ()).dimension;;
let get_dimension_management () = (get ()).dimension_management;;
let get_cutting_strategy () = (get ()).cutting_strategy;;
let get_stopping_criterion () = (get ()).stopping_criterion;;
let get_image_width () = (get ()).image_width;;
let get_image_height () = (get ()).image_height;;

(* Program configuration. *)

let get_formatter () =
  match Opt.get_out_formatter () with
  | Opt.Std_out -> Format.std_formatter
  | Opt.Std_err -> Format.err_formatter
  | Opt.Out_file fname ->
      Format.formatter_of_out_channel (Pervasives.open_out fname)
;;

let get_report_allocated_memory = Opt.get_report_allocated_memory;;

let get_incremental_mode = Opt.get_incremental_mode;;

let get_graphics_mode () =
  match Opt.get_graphics_mode () with
  | Opt.No_graphics ->
      { on = false; kind = None; }
  | Opt.Minimal_graphics ->
      { on = true; kind = Some (Param_quantity, Compound_channel); }
  | Opt.Parameter_graphics ->
      { on = true; kind = Some (Param_quantity, All_channels); }
  | Opt.Compound_graphics ->
      { on = true; kind = Some (All_quantities, Compound_channel); }
  | Opt.Full_graphics ->
      { on = true; kind = Some (All_quantities, All_channels); }
;;

let check_graphics_window_sizes () =
  let w_min = Opt.get_graphics_window_min_width ()
  and h_min = Opt.get_graphics_window_min_height () in
  let w_max = Opt.get_graphics_window_max_width ()
  and h_max = Opt.get_graphics_window_max_height () in
  if w_max < w_min then
    c_invalid_arg "Min width is greater than max width" else
  if h_max < h_min then
    c_invalid_arg "Min height is greater then max height"
;;

let get_graphics_window_min_size () =
  check_graphics_window_sizes ();
  let w_min = Opt.get_graphics_window_min_width ()
  and h_min = Opt.get_graphics_window_min_height () in
  (Nat.of_int w_min, Nat.of_int h_min)
;;

let get_graphics_window_max_size () =
  check_graphics_window_sizes ();
  let w_max = Opt.get_graphics_window_max_width ()
  and h_max = Opt.get_graphics_window_max_height () in
  (Nat.of_int w_max, Nat.of_int h_max)
;;

let get_graphics_gradient_threshold () =
  Nat.of_int (Opt.get_graphics_gradient_threshold ())
;;

let get_final_image_filename () = Opt.get_final_image_filename ();;

(* Monitoring. *)

let { Udebug.report_action = report_action;
      debug_action = debug_action; _; } =
  Options.printing_facilities
;;

let printer_ident_value ident printer _ctx ppf value =
  Format.fprintf ppf "%s =@ %a"
    ident
    (printer Og_precedence.Ck_toplevel) value
;;

let print_report printer ppf value =
  report_action
    (fun () ->
      Format.fprintf ppf "@[<hv>%a@]"
        (printer Og_precedence.Ck_toplevel) value;
      "")
;;

let print_debug fun_name ident printer ppf value =
  debug_action (Some fun_name)
    (fun () ->
      Format.fprintf ppf "@[<v 2>%s: %a@]@."
        fun_name
        (printer_ident_value ident printer Og_precedence.Ck_toplevel) value;
      "")
;;

let print_debug_no_arg fun_name printer ppf =
  debug_action (Some fun_name)
    (fun () -> Format.fprintf ppf "@[<v 2>%s: %t@]@." fun_name printer; "")
;;
