(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2004-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: ucommand_option.mli,v 1.67 2017/05/31 15:24:06 fclement Exp $ *)

(** {3 Handle options for the command line of a program}

 Regular command options are the association of an option name to use on
 the command line, an action to be executed when parsing the command line,
 and a documentation to be printed for help.
 Generalized command options have the additional capability to specify
 synonyms to the option name.

 Command options are classified either as (regular) application, application
 warn error, report, debug and version, (regular) library, library warn
 error, report, debug and version, and service command options.
 This classification drives the appearance of option documentations in the
 help message:
 application command options come first (regular, warn error, report, debug,
 then version),
 then library command options (regular, warn error, report, debug, then
 version),
 and finally service command options.

 Application command options are devoted to standalone programs.
 Library command options are devoted to libraries.

 Warn error command options are used to make execution stop on warnings.

 Report command options are used to trigger the collection and printing of
 information during the execution of code of a program or a library (e.g. to
 monitor algorithms).

 Debug command options are used to trigger the trace of the execution of code
 of a program or a library when its behavior is not satisfactory.

 Version command options are used to display version information, and exit.

 Service command options are automatically provided; they are used to trigger
 either all warn error, report, debug or version command options, or the
 printing of the help message that contains the documentation of all the
 command options (and then exit).
 Option names [-warn-error] (with aliases [-we] and [--warn-error]),
 [-report] (with aliases [-r] and [--report]),
 [-debug] (with aliases [-d] and [--debug]),
 [-long-version-info] (with aliases [-lv] and [--long-version-info]),
 [-version-info] (with aliases [-v] and [--version-info]),
 [-short-version-info] (with aliases [-sv] and [--short-version-info]),
 and [-help] (with aliases [-h] and [--help])
 are automatically provided (for each class containing at least two declared
 options).

 The parsing of command options from the command line may be split into
 phases, see the dedicated section.

 This module is implemented over module Arg of the OCaml standard library.
 Module {!Uaref} (from the OCaml user's additional library Ulib) provides a
 collection of abstract references that may be used to define the actions
 to be executed when parsing the command line; and module {!Uoption_key}
 (from Ulib as well) provides operations on option keys.
 Module {!Udebug} (also from Ulib) provides shortcuts for the definition of
 reporting and degugging facilities, as well as error message printing.
 Module {!Upph_id_card} (also from Ulib) provides shortcuts for the
 definition of version facilities. *)

(** {6 Defining command options} *)

type regular_command_option =
  {
   key : Uoption_key.t;
   spec : Arg.spec;
   doc : Arg.doc;
  }
(** The type of regular command options.
 For a given regular command option [opt], [opt.key] is the option name
 (a string starting with a dash (['-']) character), [opt.spec] is the action,
 and [opt.doc] is the documentation (a string).
 This type is equivalent to the [key * spec * doc] triples used in module
 Arg of the standard library. *)
;;

type t
and command_option = t
(** The abstract type for command options.
 The notion of command option generalizes the notion of regular command
 option by adding the possibility to specify aliases to the main option name.
 Aliases maybe of three kinds: GNU option key aliases are supposed to be long
 descriptive option names starting with two dash (['-']) characters;
 short option key aliases are supposed to be short option names with a small
 number of characters (ideally one or two, plus the initial dash (['-'])
 character);
 other option key aliases have free format.
 GNU option key and short option key aliases may be computed automatically
 from the main option name (which is then supposed to be a long descriptive
 option name). *)
;;

val of_regular : regular_command_option -> t
(** [of_regular rco] returns the command option with [rco.key] as the main
 option name, the same action and documentation, and with no option key
 alias. *)
;;

val with_gnu_option_key : Uoption_key.t -> t -> t
(** [with_gnu_option_key k co] returns the same command option with its GNU
 option key set to [k]. *)
;;

val with_short_option_key : Uoption_key.t -> t -> t
(** [with_short_option_key k co] returns the same command option with its
 short option key set to [k]. *)
;;

val with_option_key_aliases : Uoption_key.t list -> t -> t
(** [with_short_option_key kl co] returns the same command option with the
 elements of the list [kl] as option key aliases. *)
;;

val with_auto_gnu_option_key : t -> t
(** [with_auto_gnu_option_key co] returns the same command option with its
 GNU option key automatically set to its main option key prefixed by an
 additional dash (['-']) character. *)
;;

val with_auto_short_option_key : t -> t
(** [with_auto_short_option_key co] returns the same command option with its
 short option key automatically set to the concatenation of initials of words
 of its main option key prefixed by a dash (['-']) character.
 Words in the main option key are supposed to be separated by dash (['-'])
 characters. *)
;;

(** {6 Managing command options} *)

val application_add : t -> unit;;
val library_add : t -> unit
(** Add the command option given in argument in the corresponding class of
 command options.
 For each class, the order of specification in the code of the program is
 preserved in the help message. *)
;;

val application_add_all : t list -> unit;;
val library_add_all : t list -> unit
(** Same as [*_add] functions above, but add all the elements of the list of
 command options given in argument. *)
;;

val clear_all : unit -> unit
(** Clear all already added entries of all classes of command options. *)
;;

val application_clear : unit -> unit;;
val library_clear : unit -> unit
(** Clear already added entries in the corresponding class of command
 options. *)
;;

val application_clear_option_key : Uoption_key.t -> unit;;
val library_clear_option_key : Uoption_key.t -> unit
(** Clear all already added entries in the corresponding class of command
 options which main option name matches the given [key]. *)
;;

(** {6 Parsing command options} *)

type command_line = string array
(** The type of command lines. *)
;;

val set_anon_fun : Arg.anon_fun -> unit
(** Set the function called on anonymous arguments.
 Default value is [ignore]. *)
;;

val set_usage_msg : Arg.usage_msg -> unit
(** Set the usage message displayed by [-help] option.
 Default value is ["\nUsage: <command> [options]\n\nOptions are:"] where
 [<command>] is the command name used to invoke the program. *)
;;

val parse : unit -> unit
(** [parse ()] parses the {e command line to parse} (initialized to [Sys.argv]
 at the beginning of the execution) to treat command options.
 Anonymous function and usage message are those set by {!set_anon_fun} and
 {!set_usage_msg} (See module [Arg] of the standard library).

 As a side effect, the command line to parse is reduced to
 [ [| Sys.argv.(0); |] ]; thus further applications of function [parse] would
 do nothing. *)
;;

val parse_argv : command_line -> unit
(** [parse_argv argv] parses the command line [argv] to treat command options.
 See {!parse} above.

 Raise [Arg.Bad] with the error message as argument when an error occurs.
 Raise [Arg.Help] with the help message as argument when help option is given
 (either [-help], [-h], or [--help]). *)
;;

(** {6 Phased parsing of command options} *)

(** Application and library command options may be grouped into phases;
 then, command options may be parsed successively phase by phase.
 This way, some mandatory options may be processed first to be able to further
 process the remaining options.
 For instance, this may be used to obtain an early load of the configuration
 of an iterative algorithm when proceeding to a warm restart.

 To add {e phased options} to phase named ["foobar"], use either {!phase_add},
 or {!phase_add_all}, with [Phase "foobar"] as first argument.
 Of course, phased options of a library should be clearly documented to allow
 for a proper phased parsing by applications linked to it;
 in particular, the name of the phases should be clearly mentioned.

 To process all phased options of phase ["foobar"] in the command line
 to parse (initialized to [Sys.argv] at the beginning of the execution)
 (resp. [argv]), use function {!parse_phase} (resp. {!parse_phase_argv}),
 with [Phase "foobar"] as first argument.
 The remaining arguments in the command line to parse (resp. [argv]) may be
 further parsed either again with {!parse_phase} (resp. {!parse_phase_argv})
 to process options of another phase, or with {!parse} (resp. {!parse_argv})
 to process the whole remainder.

 Note that the process of anonymous arguments cannot be phased, thus anonymous
 arguments can only be parsed with {!parse} and {!parse_argv} functions. *)

type phase_name = Phase of string;;

val phase_add : phase_name -> Uoption_key.t -> unit
(** Add the command option associated to the option key argument to the
 specified phase.
 Option key aliases are automatically taken into account.

 Raise [Failure] if the option key is not recorded in any of the command
 option classes. *)
;;

val phase_add_all : phase_name -> Uoption_key.t list -> unit
(** Repeat {!phase_add} for all option keys of the argument list.

 Raise [Failure] if any of the option keys is not recorded in any of the
 command option classes. *)
;;

val phase_clear : phase_name -> Uoption_key.t -> unit
(** Remove the command option associated to the option key argument from the
 specified phase.

 Raise [Failure] if the phase is undefined. *)
;;

val parse_phase : phase_name -> unit
(** [parse_phase phase_name] parses the {e command line to parse}
 (initialized to [Sys.argv] at the beginning of the execution) to treat
 phased options from phase [phase_name].

 As a side effect, matching command options are suppressed from the command
 line to parse; and the remaining arguments may be parsed further using
 either {!parse}, or {!parse_phase} again. *)
;;

val parse_phase_argv : phase_name -> command_line -> command_line
(** [parse_phase_argv phase_name argv] parses the command line [argv] to
 treat phased options from phase [phase_name], and returns the remaining
 command line arguments.
 Those remaining command line arguments may be parsed further using either
 {!parse_argv}, or {!parse_phase_argv} again.

 Raise [Arg.Bad] with the error message as argument when an error occurs.
 Raise [Arg.Help] with the help message as argument when help option is given
 (either [-help], [-h], or [--help]). *)
;;

(** {6 Useful shortcuts} *)

val application_add_all_rco : regular_command_option list -> unit;;
val library_add_all_rco : regular_command_option list -> unit
(** Same as [*_add_all] functions above with no additional option key
 aliases. *)
;;

val application_add_all_rco_auto : regular_command_option list -> unit;;
val library_add_all_rco_auto : regular_command_option list -> unit
(** Same as [*_add_all_rco] functions above with automatic addition of GNU
 option key and short option key aliases. *)
;;

val application_add_all_rco_with_sk :
  (Uoption_key.short_option_key_status * regular_command_option) list -> unit
;;
val library_add_all_rco_with_sk :
  (Uoption_key.short_option_key_status * regular_command_option) list -> unit
(** Same as [*_add_all_rco] functions above, but the
 {!Uoption_key.short_option_key_status} is specified: either automatic or
 given. *)
;;

val parse_all_services : unit -> unit
(** [parse_all_services ()] parses the {e command line to parse} (initialized
 to [Sys.argv] at the beginning of the execution) to treat all services
 (help/versioning/debugging).

 As a side effect, matching command options are suppressed from the command
 line to parse; and the remaining arguments may be parsed further using
 either {!parse}, or {!parse_phase}. *)
;;

val parse_all_services_argv : command_line -> command_line
(** [parse_all_services_argv argv] parses the command line [argv] to treat
 all services (help/versioning/debugging), and returns the remaining command
 line arguments.
 Those remaining command line arguments may be parsed further using either
 {!parse_argv}, or {!parse_phase_argv}.

 Raise [Arg.Bad] with the error message as argument when an error occurs.
 Raise [Arg.Help] with the help message as argument when help option is given
 (either [-help], [-h], or [--help]). *)
;;

(**/**)
(** {6 Undocumented features} *)

type 'a option_class =
  | Regular of 'a
  | Warn_error of 'a
  | Report of 'a
  | Debug of 'a
  | Long_version_info of 'a
  | Version_info of 'a
  | Short_version_info of 'a
  | Service
(** The polymorphic type of option classes. *)
;;

val string_of_option_kind : unit option_class -> string;;
val option_keys_of_option_kind :
  unit option_class -> Uoption_key.option_keys
;;

type software_purpose =
  | Application
  | Library
(** The purpose of a piece of software is either and application, or a
 library that will eventually be linked with some application by some
 software developer. *)
;;

type software_option_key_data =
  {
   name : Upervasives.software_name;
   option_key_prefixes_opt : Uoption_key.option_keys option;
  }
(** Data to build automatic service command option in modules {!Udebug},
 and {!Upph_id_card}. *)
;;

val warn_error_app_add : t -> unit;;
val warn_error_lib_add : t -> unit;;
val report_app_add : t -> unit;;
val report_lib_add : t -> unit;;
val debug_app_add : t -> unit;;
val debug_lib_add : t -> unit;;
val long_version_info_app_add : t -> unit;;
val long_version_info_lib_add : t -> unit;;
val version_info_app_add : t -> unit;;
val version_info_lib_add : t -> unit;;
val short_version_info_app_add : t -> unit;;
val short_version_info_lib_add : t -> unit
(** Same as [*_add] functions above, but prefer shortcuts provided by modules
 {!Udebug} and {!Upph_id_card}. *)
;;

val warn_error_app_add_all : t list -> unit;;
val warn_error_lib_add_all : t list -> unit;;
val report_app_add_all : t list -> unit;;
val report_lib_add_all : t list -> unit;;
val debug_app_add_all : t list -> unit;;
val debug_lib_add_all : t list -> unit;;
val long_version_info_app_add_all : t list -> unit;;
val long_version_info_lib_add_all : t list -> unit;;
val version_info_app_add_all : t list -> unit;;
val version_info_lib_add_all : t list -> unit;;
val short_version_info_app_add_all : t list -> unit;;
val short_version_info_lib_add_all : t list -> unit
(** Same as functions [*_add_all] above, but prefer shortcuts provided by
 modules {!Udebug} and {!Upph_id_card}. *)
;;

val warn_error_app_clear : unit -> unit;;
val warn_error_lib_clear : unit -> unit;;
val report_app_clear : unit -> unit;;
val report_lib_clear : unit -> unit;;
val debug_app_clear : unit -> unit;;
val debug_lib_clear : unit -> unit;;
val long_version_info_app_clear : unit -> unit;;
val long_version_info_lib_clear : unit -> unit;;
val version_info_app_clear : unit -> unit;;
val version_info_lib_clear : unit -> unit;;
val short_version_info_app_clear : unit -> unit;;
val short_version_info_lib_clear : unit -> unit
(** Same as functions [*_clear] above, but prefer shortcuts provided by
 modules {!Udebug} and {!Upph_id_card}. *)
;;

val warn_error_app_clear_option_key : Uoption_key.t -> unit;;
val warn_error_lib_clear_option_key : Uoption_key.t -> unit;;
val report_app_clear_option_key : Uoption_key.t -> unit;;
val report_lib_clear_option_key : Uoption_key.t -> unit;;
val debug_app_clear_option_key : Uoption_key.t -> unit;;
val debug_lib_clear_option_key : Uoption_key.t -> unit;;
val long_version_info_app_clear_option_key : Uoption_key.t -> unit;;
val long_version_info_lib_clear_option_key : Uoption_key.t -> unit;;
val version_info_app_clear_option_key : Uoption_key.t -> unit;;
val version_info_lib_clear_option_key : Uoption_key.t -> unit;;
val short_version_info_app_clear_option_key : Uoption_key.t -> unit;;
val short_version_info_lib_clear_option_key : Uoption_key.t -> unit
(** Same as functions [*_clearkey] above, but prefer shortcuts provided by
 modules {!Udebug} and {!Upph_id_card}. *)
;;

val warn_error_app_add_all_rco : regular_command_option list -> unit;;
val warn_error_lib_add_all_rco : regular_command_option list -> unit;;
val report_app_add_all_rco : regular_command_option list -> unit;;
val report_lib_add_all_rco : regular_command_option list -> unit;;
val debug_app_add_all_rco : regular_command_option list -> unit;;
val debug_lib_add_all_rco : regular_command_option list -> unit;;
val long_version_info_app_add_all_rco : regular_command_option list -> unit;;
val long_version_info_lib_add_all_rco : regular_command_option list -> unit;;
val version_info_app_add_all_rco : regular_command_option list -> unit;;
val version_info_lib_add_all_rco : regular_command_option list -> unit;;
val short_version_info_app_add_all_rco : regular_command_option list -> unit;;
val short_version_info_lib_add_all_rco : regular_command_option list -> unit
(** Same as functions [*_add_all_rco] above, but prefer shortcuts
 provided by modules {!Udebug} and {!Upph_id_card}. *)
;;

val warn_error_app_add_all_rco_auto : regular_command_option list -> unit;;
val warn_error_lib_add_all_rco_auto : regular_command_option list -> unit;;
val report_app_add_all_rco_auto : regular_command_option list -> unit;;
val report_lib_add_all_rco_auto : regular_command_option list -> unit;;
val debug_app_add_all_rco_auto : regular_command_option list -> unit;;
val debug_lib_add_all_rco_auto : regular_command_option list -> unit;;
val long_version_info_app_add_all_rco_auto :
  regular_command_option list -> unit
;;
val long_version_info_lib_add_all_rco_auto :
  regular_command_option list -> unit
;;
val version_info_app_add_all_rco_auto :
  regular_command_option list -> unit
;;
val version_info_lib_add_all_rco_auto :
  regular_command_option list -> unit
;;
val short_version_info_app_add_all_rco_auto :
  regular_command_option list -> unit
;;
val short_version_info_lib_add_all_rco_auto :
  regular_command_option list -> unit
(** Same as functions [*_add_all_rco_auto] above, but prefer shortcuts
 provided by modules {!Udebug} and {!Upph_id_card}. *)
;;

val warn_error_app_add_all_rco_with_sk :
  (Uoption_key.short_option_key_status * regular_command_option) list -> unit
;;
val warn_error_lib_add_all_rco_with_sk :
  (Uoption_key.short_option_key_status * regular_command_option) list -> unit
;;
val report_app_add_all_rco_with_sk :
  (Uoption_key.short_option_key_status * regular_command_option) list -> unit
;;
val report_lib_add_all_rco_with_sk :
  (Uoption_key.short_option_key_status * regular_command_option) list -> unit
;;
val debug_app_add_all_rco_with_sk :
  (Uoption_key.short_option_key_status * regular_command_option) list -> unit
;;
val debug_lib_add_all_rco_with_sk :
  (Uoption_key.short_option_key_status * regular_command_option) list -> unit
;;
val long_version_info_app_add_all_rco_with_sk :
  (Uoption_key.short_option_key_status * regular_command_option) list -> unit
;;
val long_version_info_lib_add_all_rco_with_sk :
  (Uoption_key.short_option_key_status * regular_command_option) list -> unit
;;
val version_info_app_add_all_rco_with_sk :
  (Uoption_key.short_option_key_status * regular_command_option) list -> unit
;;
val version_info_lib_add_all_rco_with_sk :
  (Uoption_key.short_option_key_status * regular_command_option) list -> unit
;;
val short_version_info_app_add_all_rco_with_sk :
  (Uoption_key.short_option_key_status * regular_command_option) list -> unit
;;
val short_version_info_lib_add_all_rco_with_sk :
  (Uoption_key.short_option_key_status * regular_command_option) list -> unit
(** Same as functions [*_add_all_rco_with_sk] above,
 but prefer shortcuts provided by modules {!Udebug} and {!Upph_id_card}. *)
;;

val phase_ulib_debugging : phase_name
(** Ulib debugging phase name. *)
;;

val phase_warn_error : unit -> unit;;
val phase_report : unit -> unit;;
val phase_debug : unit -> unit
(** Add all declared command options of the underlying class to the phase
 {!phase_ulib_debugging}, including automatically added service command
 option. Should only be used when all command options are declared, just
 before parsing the command line. *)
;;

val phase_all_debugging : unit -> unit
(** Shortcut to call all three [phase_*] functions above. *)
;;

val phase_ulib_versioning : phase_name
(** Ulib versioning phase name. *)
;;

val phase_long_version_info : unit -> unit;;
val phase_version_info : unit -> unit;;
val phase_short_version_info : unit -> unit
(** Add all declared command options of the underlying class to the phase
 {!phase_ulib_versioning}, including automatically added service command
 option. Should only be used when all command options are declared, just
 before parsing the command line. *)
;;

val phase_all_versioning : unit -> unit
(** Shortcut to call all three [phase_*] functions above. *)
;;

val phase_ulib_helping : phase_name
(** Ulib helping phase name. *)
;;

val phase_help : unit -> unit
(** Add help command option to the phase {!phase_ulib_helping}. Should only
 be used when all command options are declared, just before parsing the
 command line. *)
;;

val phase_all_services : unit -> unit
(** Shortcut to call all two [phase_all_*] and {!phase_help} functions
  above. *)
;;

val parse_ulib_helping_argv : command_line -> command_line;;
val parse_ulib_versioning_argv : command_line -> command_line;;
val parse_ulib_debugging_argv : command_line -> command_line;;

val parse_ulib_helping : unit -> unit;;
val parse_ulib_versioning : unit -> unit;;
val parse_ulib_debugging : unit -> unit;;

(**/**)
(** {6 For tests only} *)

val arg_catch : ('a -> 'b) -> 'a -> 'b;;
