(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2004-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: ulistn.ml,v 1.18 2017/05/31 15:24:06 fclement Exp $ *)

module Nat = Unatural;;
module L = Ulist;;

let _0_ = Nat._0_;;

let qual = Upervasives.qualify "Ulistn";;
let invalid_arg = Upervasives.invalid_arg_qual qual;;

open Ulistn_spec;;

module Succ_structure =
  functor (Lpns : LISTN_STRUCTURE) ->
struct
  type length = Upervasives.length
  and index = Upervasives.index
  ;;

  let min_length = Nat.succ Lpns.min_length;;

  type 'a listpn = 'a Lpns.t;;
  type 'a listn = 'a * 'a listpn;;
  type 'a t = 'a listn;;

  let to_listpn (h, lpn) = Lpns.cons h lpn;;
  let of_listpn lpn = (Lpns.hd lpn, Lpns.of_listpn (Lpns.tl lpn));;

  let to_list (h, lpn) = h :: Lpns.to_list lpn;;
  let of_list = function
    | [] -> invalid_arg "of_list" "empty list"
    | h :: lpn -> (h, Lpns.of_list lpn)
  ;;

  let cons x (h, lpn) = (x, Lpns.cons h lpn);;
  let hd (h, _) = h;;
  let tl (_, l) = l;;
end
;;

module Make =
  functor (Lns : LISTN_STRUCTURE) ->
struct
  include Lns;;

  let lift_dom f ln = f (to_list ln);;
  let lift_codom f x = of_list (f x);;
  let lift f = lift_codom (lift_dom f);;
  let lift_dom2 f ln1 ln2 = f (to_list ln1) (to_list ln2);;
  let lift_codom2 f x1 x2 = of_list (f x1 x2);;
  let lift2 f = lift_codom2 (lift_dom2 f);;

  let proj_codom f x = to_list (f x);;
  let proj_codom_i f i = proj_codom (f i);;
  let proj_codom2 f x1 x2 = to_list (f x1 x2);;
  let proj_codom2_i f i = proj_codom2 (f i);;

  let length ln = lift_dom L.length ln;;
  let first_p n ln = lift_dom (L.first_p n) ln;;
  let last_p n ln = lift_dom (L.last_p n) ln;;
  let last ln = lift_dom L.last ln;;
  let nth ln n = lift_dom (fun l -> L.nth l n) ln;;
  let rank e = lift_dom (L.rank e);;
  let rev ln = lift L.rev ln;;
  let append ln1 ln2 = lift2 L.append ln1 ln2;;
  let rev_append ln1 ln2 = lift2 L.rev_append ln1 ln2;;
  let init n f = of_list (L.init n f);;
  let rev_init n f = of_list (L.rev_init n f);;
  let replace i x = lift (L.replace i x);;
  let rev_replace i x = lift (L.rev_replace i x);;
  let replace_by_list i = lift2 (L.replace_by_list i);;
  let rev_replace_by_list i = lift2 (L.rev_replace_by_list i);;
  let insert i = lift2 (L.insert i);;
  let rev_insert i = lift2 (L.rev_insert i);;

  (* Iterators. *)
  let iter f = lift_dom (L.iter f);;
  let iter_i_shift f i = lift_dom (L.iter_i_shift f i);;
  let iter_i f = iter_i_shift f _0_;;
  let map f = lift (L.map f);;
  let map_i_shift f i = lift (L.map_i_shift f i);;
  let map_i f = map_i_shift f _0_;;
  let rev_map f = lift (L.rev_map f);;
  let rev_map_i_end_shift accu f i =
    lift (L.rev_map_i_end_shift accu f i)
  ;;
  let rev_map_i f = rev_map_i_end_shift [] f _0_;;
  let flat_map_end accu f = lift (L.flat_map_end accu (proj_codom f));;
  let flat_map f = flat_map_end [] f;;
  let concat ll = flat_map Upervasives.id ll;;
  let flatten = concat;;
  let flat_map_i_end_shift accu f i =
    lift (L.flat_map_i_end_shift accu (proj_codom_i f) i)
  ;;
  let flat_map_i f = flat_map_i_end_shift [] f _0_;;
  let rev_flat_map_end accu f =
    lift (L.rev_flat_map_end accu (proj_codom f))
  ;;
  let rev_flat_map f = rev_flat_map_end [] f;;
  let rev_flat_map_i_end_shift accu f i =
    lift (L.rev_flat_map_i_end_shift accu (proj_codom_i f) i)
  ;;
  let rev_flat_map_i f = rev_flat_map_i_end_shift [] f _0_;;
  let fold_left f accu = lift_dom (L.fold_left f accu);;
  let fold_left_i_shift f i accu =
    lift_dom (L.fold_left_i_shift f i accu)
  ;;
  let fold_left_i f = fold_left_i_shift f _0_;;
  let fold_right f ln accu =
    lift_dom (fun l -> L.fold_right f l accu) ln
  ;;
  let fold_right_i_shift f i ln accu =
    lift_dom (fun l -> L.fold_right_i_shift f i l accu) ln
  ;;
  let fold_right_i f = fold_right_i_shift f _0_;;
  let rev_fold_left_env f (env, accu) ln =
    let f (env, accu) x =
      let (env, accu) = f (env, accu) x in
      (env, to_list accu) in
    let (env, accu) = lift_dom (L.fold_left f (env, accu)) ln in
    (env, of_list accu)
  ;;
  let fold_left_env f (env, accu) ln =
    let (env, accu) = rev_fold_left_env f (env, L.rev accu) ln in
    (env, rev accu)
  ;;
  let fold_left_env_init f env = fold_left_env f (env, []);;

  (* Iterators on two lists with minimum length. *)
  let iter2 f = lift_dom2 (L.iter2 f);;
  let iter2_i_shift f i = lift_dom2 (L.iter2_i_shift f i);;
  let iter2_i f = iter2_i_shift f _0_;;
  let map2 f = lift2 (L.map2 f);;
  let map2_i_shift f i = lift2 (L.map2_i_shift f i);;
  let map2_i f = map2_i_shift f _0_;;
  let rev_map2 f = lift2 (L.rev_map2 f);;
  let rev_map2_i_end_shift accu f i =
    lift2 (L.rev_map2_i_end_shift accu f i)
  ;;
  let rev_map2_i f = rev_map2_i_end_shift [] f _0_;;
  let flat_map2_end accu f = lift2 (L.flat_map2_end accu (proj_codom2 f));;
  let flat_map2 f = flat_map2_end [] f;;
  let flat_map2_i_end_shift accu f i =
    lift2 (L.flat_map2_i_end_shift accu (proj_codom2_i f) i)
  ;;
  let flat_map2_i f = flat_map2_i_end_shift [] f _0_;;
  let rev_flat_map2_end accu f =
    lift2 (L.rev_flat_map2_end accu (proj_codom2 f))
  ;;
  let rev_flat_map2 f = rev_flat_map2_end [] f;;
  let rev_flat_map2_i_end_shift accu f i =
    lift2 (L.rev_flat_map2_i_end_shift accu (proj_codom2_i f) i)
  ;;
  let rev_flat_map2_i f = rev_flat_map2_i_end_shift [] f _0_;;
  let fold_left2 f accu = lift_dom2 (L.fold_left2 f accu);;
  let fold_left2_i_shift f i accu =
    lift_dom2 (L.fold_left2_i_shift f i accu)
  ;;
  let fold_left2_i f = fold_left2_i_shift f _0_;;
  let fold_right2 f ln1 ln2 accu =
    lift_dom2 (fun l1 l2 -> L.fold_right2 f l1 l2 accu) ln1 ln2
  ;;
  let fold_right2_i_shift f i ln1 ln2 accu =
    lift_dom2 (fun l1 l2 -> L.fold_right2_i_shift f i l1 l2 accu) ln1 ln2
  ;;
  let fold_right2_i f = fold_right2_i_shift f _0_;;
  let rev_fold_left2_env f (env, accu) ln1 ln2 =
    let f (env, accu) x1 x2 =
      let (env, accu) = f (env, accu) x1 x2 in
      (env, to_list accu) in
    let (env, accu) = lift_dom2 (L.fold_left2 f (env, accu)) ln1 ln2 in
    (env, of_list accu)
  ;;
  let fold_left2_env f (env, accu) lpn1 lpn2 =
    let (env, accu) = rev_fold_left2_env f (env, L.rev accu) lpn1 lpn2 in
    (env, rev accu)
  ;;
  let fold_left2_env_init f env = fold_left2_env f (env, []);;

  (* Scanning. *)
  let for_all p = lift_dom (L.for_all p);;
  let exists p = lift_dom (L.exists p);;
  let for_all2 p = lift_dom2 (L.for_all2 p);;
  let exists2 p = lift_dom2 (L.exists2 p);;
  let mem a = exists (fun x -> x = a);;
  let memq a = exists (fun x -> x == a);;

  (* Searching. *)
  let find p = lift_dom (L.find p);;
  let find_all p = lift_dom (L.find_all p);;
  let filter = find_all;;
  let partition p = lift_dom (L.partition p);;

  (* Association lists with minimum length. *)
  let assoc k = lift_dom (L.assoc k);;
  let assq k = lift_dom (L.assq k);;
  let mem_assoc k = exists (fun (a, _) -> k = a);;
  let mem_assq k = exists (fun (a, _) -> k == a);;
  let remove_assoc k = lift_dom (L.remove_assoc k);;
  let remove_assq k = lift_dom (L.remove_assq k);;
  let compare_assoc cmp = lift_dom (L.compare_assoc cmp);;
  let minus_pervasives_compare x y = - Pervasives.compare x y;;
  let min_assoc ln = compare_assoc Pervasives.compare ln
  and max_assoc ln = compare_assoc minus_pervasives_compare ln
  ;;
  let compare_fst_assoc = compare_assoc
  and min_fst_assoc = min_assoc
  and max_fst_assoc = max_assoc
  ;;
  let compare_snd_assoc cmp = lift_dom (L.compare_snd_assoc cmp);;
  let min_snd_assoc ln = compare_snd_assoc Pervasives.compare ln
  and max_snd_assoc ln = compare_snd_assoc minus_pervasives_compare ln
  ;;

  (* Lists of pairs with minimum length. *)
  let lift_pair f ln =
    let (res1, res2) = lift_dom f ln in
    (of_list res1, of_list res2)
  ;;
  let split ln = lift_pair L.split ln;;
  let combine ln1 ln2 = lift2 L.combine ln1 ln2;;

  (* Sorting. *)
  let sort cmp = lift (L.sort cmp);;
  let stable_sort cmp = lift (L.stable_sort cmp);;
  let fast_sort cmp = lift (L.fast_sort cmp);;
  let merge cmp = lift2 (L.merge cmp);;
  let of_range pred l r = of_list (L.of_range pred l r);;
end
;;

module Succ = functor (Ln : LISTN) -> Make (Succ_structure (Ln));;
