(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2004-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: unonempty_set.ml,v 1.13 2017/05/31 15:24:06 fclement Exp $ *)

(* Nonempty sets. *)

module Nat = Unatural;;
module L1 = Ulist1;;
module A = Uarray;;
module Idx = Uindex;;
module Univ = Uuniverse;;
module Nst = Unonempty_set_types;;

let _0_ = Nat._0_;;

let invalid_arg = Nst.invalid_arg
and ucheck_index = Nst.ucheck_index
;;

type 'a uindex = 'a Nst.uindex;;

type 'a index = 'a Nst.index;;

type 'a t = 'a Nst.t = private
  {
   universe : 'a Univ.t;
   uindices : 'a uindex Univ.t;
  }
and 'a nonempty_set = 'a t
;;

(* Construction functions. *)

let full = Nst.full
let of_array = Nst.of_array
and of_list = Nst.of_list
and unsafe_of_list = Nst.unsafe_of_list
and union = Nst.union
;;

let singleton u uindex = of_array u [| uindex; |];;

let of_rectangle h w = full (Univ.of_rectangle h w);;

let of_nat_rectangle h w = full (Univ.of_nat_rectangle h w);;

(* Indices and elements. *)

let cardinal = Nst.cardinal
and same_universe = Nst.same_universe
and check_same_universe = Nst.check_same_universe
;;

let valid_index nes index = Idx.to_nat index < cardinal nes;;

let check_index fun_name nes index =
  if not (valid_index nes index) then
    invalid_arg fun_name
      (Printf.sprintf "index %i out of set of %d element(s)"
         (Idx.to_int index)
         (Nat.to_int (cardinal nes)))
;;

let mem nes uindex = Univ.mem nes.uindices uindex;;

let choice nes = Univ.element nes.uindices (Idx.of_nat _0_);;

let uindex nes index =
  check_index "indice" nes index;
  Univ.element nes.uindices index
;;

let uindices nes = Univ.elements nes.uindices;;

let element nes index = Univ.element nes.universe (uindex nes index);;

let elements nes = Univ.sub_elements nes.universe (uindices nes);;

let add nes uindex =
  ucheck_index "add_to_nonempty" nes.universe uindex;
  if not (mem nes uindex) then Univ.add nes.uindices uindex
;;

(* Iterators. *)

let fold_left = Nst.fold_left;;

let fold_left_i f accu nes = Univ.fold_left_i f accu nes.uindices;;

let iter f nes = Univ.iter f nes.uindices;;

let iter_i f nes = Univ.iter_i f nes.uindices;;

let map_indices f nes = Univ.map_elements f nes.uindices;;

let map f nes u = of_array u (map_indices f nes);;

let find p nes = Univ.find p nes.uindices;;

let find_all p nes = Univ.find_all p nes.uindices;;

let filter = find_all;;

(* Nonempty subsets. *)

type 'a nonempty_subset =
  {
   support : 'a t;
   indices : 'a index array;
   nonempty_set : 'a t;
  }
;;

let is_full nes = cardinal nes = Univ.cardinal nes.universe;;

let check_full fun_name nes =
  if not (is_full nes) then invalid_arg fun_name "nonempty set is not full"
;;

let same_full nes1 nes2 =
  same_universe nes1 nes2 && is_full nes1 && is_full nes2
;;

let check_same_full fun_name nes1 nes2 =
  check_same_universe fun_name nes1 nes2;
  check_full fun_name nes1;
  check_full fun_name nes2
;;

let (full_index_of_uindex : 'a uindex -> 'a index) =
  fun uindex -> Idx.of_nat (Idx.to_nat uindex)
;;

(* We use here the fact that the full set, containing all the elements of the
  universe, is numbered in the same order as the universe.
  Hence, the use of magic function index_of_uindex.
  If efficiency is an issue, consider using Obj.magic on the whole array of
  indices. *)
let subset_of_full full nes =
  check_full "subset_of_full" full;
  check_same_universe "subset_of_full" full nes;
  let indices =
    A.map full_index_of_uindex (Univ.elements nes.uindices) in
  { support = full; indices = indices; nonempty_set = nes; }
;;

type 'a dichotomy = 'a nonempty_subset * 'a nonempty_subset;;

let dichotomy p nes =
  let dicho (tl, fl) index uindex =
    if p index uindex then (index :: tl, fl) else (tl, index :: fl) in
  let (tl, fl) =
    match fold_left_i dicho ([], []) nes with
    | ([], _) | (_, []) -> invalid_arg "dichotomy" "empty subset"
    | (tx :: tl, fx :: fl) -> ((tx, tl), (fx, fl)) in
  let uindex index = Univ.element nes.uindices index in
  let (utl, ufl) = (L1.rev_map uindex tl, L1.rev_map uindex fl) in
  let (tnes, fnes) =
    (unsafe_of_list nes.universe utl, unsafe_of_list nes.universe ufl) in
  ({ support = nes; indices = A.of_list1 tl; nonempty_set = tnes; },
   { support = nes; indices = A.of_list1 fl; nonempty_set = fnes; })
;;
