(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2004-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: uoption_key.ml,v 1.2 2017/05/31 15:24:06 fclement Exp $ *)

(* Operations on option keys. *)

module Nat = Unatural;;
module L = Ulist;;
module S = Ustring;;
module Okt = Uoption_key_types;;

let _0_ = Nat._0_;;

let qual = Upervasives.qualify "Uoption_key";;
let failwith = Upervasives.failwith_qual qual;;

type t = Okt.t
and option_key = t
;;

let of_string = Okt.to_t;;
external unsafe_of_string : string -> t = "%identity";;
external to_string : t -> string = "%identity";;

(* Long/short/gnu option keys. *)

let suppress_leading_dash ok = S.chop_prefix "-" (to_string ok);;
let add_leading_dash ok = unsafe_of_string ("-" ^ (to_string ok));;

type long_option_key = t
and short_option_key = t

and short_option_key_status =
  | Automatic
  | Short of short_option_key

and option_keys =
  {
   long : long_option_key;
   short : short_option_key_status;
  }
;;

let is_dash c = c = '-';;

let mask_non_alphanumeric mask c =
  if U2char.is_alphanumeric c then c else mask
;;

let long_option_key_of_string s =
  let mask = '-' in
  let is_mask c = c = mask in
  let s' =
    S.trim_edges is_mask
      (S.trim_consecutive is_mask
        (S.lowercase
          (S.iso_latin1_to_us_ascii (fun _ -> "")
            (S.map (mask_non_alphanumeric mask) s)))) in
  if S.length s' = _0_ then
    failwith "long_option_key_of_string"
      (Printf.sprintf "Empty long option key prefix from string %S" s) else
  of_string (Printf.sprintf "-%s" s')
;;

let short_option_key_of_long_option_key lok =
  let lok = S.trim_consecutive is_dash (to_string lok) in
  match S.indices lok '-' with
  | [] -> assert false
  | il ->
    let len = L.length il in
    let buf = Buffer.create (Upervasives.length_to_int len) in
    L.iter (fun i -> Buffer.add_char buf (S.get lok (Nat.succ i))) il;
    let sok = Printf.sprintf "-%s" (Buffer.contents buf) in
    if sok = lok then
      failwith "short_option_key_of_long_option_key"
        (Printf.sprintf
          "argument %S was already a short option key" lok) else
    of_string sok
;;

let make_option_keys sname sok_st =
  { long = long_option_key_of_string sname; short = sok_st; }
;;

let concat_long_option_key lok1 lok2 =
  unsafe_of_string (to_string lok1 ^ to_string lok2)
;;

let concat_short_option_key sok1 sok2 =
  unsafe_of_string (to_string sok1 ^ suppress_leading_dash sok2)
;;

let concat_option_keys
    { long = lok1; short = sok1_opt; }
    { long = lok2; short = sok2_opt; } =
  let sok_st =
    match (sok1_opt, sok2_opt) with
    | (Automatic, Automatic) -> Automatic
    | (Short _, _) | (_, Short _) ->
      let (sok1, sok2) =
        begin match (sok1_opt, sok2_opt) with
        | (Automatic, Automatic) -> assert false
        | (Short sok1, Short sok2) -> (sok1, sok2)
        | (Short sok1, Automatic) ->
            (sok1, short_option_key_of_long_option_key lok2)
        | (Automatic, Short sok2) ->
            (short_option_key_of_long_option_key lok1, sok2)
        end in
      Short (concat_short_option_key sok1 sok2) in
  { long = concat_long_option_key lok1 lok2; short = sok_st; }
;;
