(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2004-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: uoption_key.mli,v 1.2 2017/05/31 15:24:06 fclement Exp $ *)

(** {3 Operations on option keys} *)

type t = Uoption_key_types.t
and option_key = t
(** The type of option keys as to be used with module Arg.
 Option key strings have a leading dash (['-']) character and are made of at
 least two characters. *)
;;

val of_string : string -> t;;
external to_string : t -> string = "%identity";;

(** {6 Long/short/GNU option keys} *)

type long_option_key = t
and short_option_key = t

and short_option_key_status =
  | Automatic
  | Short of short_option_key

and option_keys =
  {
   long : long_option_key;
   short : short_option_key_status;
  }
(** The collection of types of option keys as to be used with module
 {!Ucommand_option}.
 [Automatic] short option keys are computed using function
 {!short_option_key_of_long_option_key} below. *)
;;

val add_leading_dash : t -> t
(** So called ``GNU option keys'' are traditionally made of two leading dash
 (['-']) characters. *)
;;

val long_option_key_of_string : string -> long_option_key
(** [long_option_key_of_string s] maps argument string [s] to the same string
 with a leading dash (['-']) prefix, and where all plain ascii alphabetic
 characters are replaced by their lowercase form, and all consecutive non
 plain ascii alphanumeric characters are replaced by a single dash (['-'])
 character.

 Raise [Failure] if the result would reduce to ["-"]. *)
;;

val short_option_key_of_long_option_key : long_option_key -> short_option_key
(** [short_option_key_of_long_option_key lk] returns the option key made of a
 dash (['-']) character followed by the initial of all dash-separated words
 of the argument long option key [lk].

 Raise [Failure] if the result would be equal to [lk]. *)
;;

val make_option_keys : string -> short_option_key_status -> option_keys
(** [make_option_keys s sk_st] returns the [option_keys] record where the
 [long] field contains the result of [long_option_key_of_string s], and the
 [short] field contains [sk_st]. *)
;;

val concat_option_keys : option_keys -> option_keys -> option_keys
(** [concat_option_keys ks1 ks2] returns the [option_keys] record where the
 [long] field contains [ks1.long ^ ks2.long], and the [short] field contains
 either [Automatic] when both [ks1.short] and [ks2.short] equal [Automatic],
 or [sk1 ^ sk2'] where [ski] is the effective short option key corresponding
 to [ksi] and [sk2'] is [sk2] without its leading dash (['-']) character
 otherwise. *)
;;

(**/**)
(** {6 Undocumented features} *)

external unsafe_of_string : string -> t = "%identity";;
