(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2004-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: upph_id_card.ml,v 1.29 2017/05/31 15:24:06 fclement Exp $ *)

(* Handling the user part of the id card of a Pph project. *)

module K = Uoption_key;;
module Co = Ucommand_option;;
module D = Udebug;;
module V = Uversion;;

type t = {
  project_name : string;
  project_banner_title : string;
  project_outline : string;
  project_release_date : string;
  project_copyright_holders : string;
  project_authors : string;
  project_http_url : string;
  project_ftp_url : string;
  project_release_ident : string;
  project_distribution_ident : string;
  project_version : string;
  project_version_full_info : string;
  project_release_name : string;
  project_distribution_name : string;
}
and pph_id_card = t
;;

(* Printing. *)

let lv_msg pic =
  Printf.sprintf "%s, %s, version %s\n"
    pic.project_name
    pic.project_banner_title
    pic.project_version_full_info

and v_msg pic =
  Printf.sprintf "%s, version %s\n"
    pic.project_name
    pic.project_release_ident

and sv_msg pic =
  Printf.sprintf "%s\n"
    pic.project_version
;;

let prerr = Printf.eprintf "%s";;

let print_long_version_info pic = prerr (lv_msg pic)
and print_version_info pic = prerr (v_msg pic)
and print_short_version_info pic = prerr (sv_msg pic)
;;

(* Command options. *)

type option_kind =
  | Long_version_info
  | Version_info
  | Short_version_info
;;

let option_class_of_option_king = function
  | Long_version_info -> Co.Long_version_info ()
  | Version_info -> Co.Version_info ()
  | Short_version_info -> Co.Short_version_info ()
;;

let option_keys_of_option_kind ok =
  Co.option_keys_of_option_kind (option_class_of_option_king ok)
;;

let gen_doc info pic =
  Printf.sprintf
    "\n\tPrint %s%s version information and exit."
    pic.project_name
    info
;;

let lv_doc = gen_doc " detailed"
and v_doc = gen_doc ""
and sv_doc = gen_doc " compact"
;;

let lv_keys = option_keys_of_option_kind Long_version_info
and v_keys = option_keys_of_option_kind Version_info
and sv_keys = option_keys_of_option_kind Short_version_info
;;

let gen_add purpose pic kps_opt =
  let (lv_keys, v_keys, sv_keys) =
    match kps_opt with
    | None -> (lv_keys, v_keys, sv_keys)
    | Some kps ->
      (K.concat_option_keys kps lv_keys,
       K.concat_option_keys kps v_keys,
       K.concat_option_keys kps sv_keys) in
  match purpose with
  | Co.Application ->
    V.add_app_long_version_option lv_keys (lv_msg pic) (lv_doc pic);
    V.add_app_version_option v_keys (v_msg pic) (v_doc pic);
    V.add_app_short_version_option sv_keys (sv_msg pic) (sv_doc pic);
  | Co.Library ->
    V.add_lib_long_version_option lv_keys (lv_msg pic) (lv_doc pic);
    V.add_lib_version_option v_keys (v_msg pic) (v_doc pic);
    V.add_lib_short_version_option sv_keys (sv_msg pic) (sv_doc pic);
;;

let add_app_version_options = gen_add Co.Application
and add_lib_version_options = gen_add Co.Library
;;

let make_option_keys pic = K.make_option_keys pic.project_name;;

let add_app_version_options_with_sk pic sk_st =
  let kps = make_option_keys pic sk_st in
  add_app_version_options pic (Some kps)

and add_lib_version_options_with_sk pic sk_st =
  let kps = make_option_keys pic sk_st in
  add_lib_version_options pic (Some kps)
;;

(* Managing automatic service options. *)

let long_version_info_app_clear = Co.long_version_info_app_clear
and long_version_info_lib_clear = Co.long_version_info_lib_clear
and version_info_app_clear = Co.version_info_app_clear
and version_info_lib_clear = Co.version_info_lib_clear
and short_version_info_app_clear = Co.short_version_info_app_clear
and short_version_info_lib_clear = Co.short_version_info_lib_clear
;;

let long_version_info_app_clear_option_key =
  Co.long_version_info_app_clear_option_key
and long_version_info_lib_clear_option_key =
  Co.long_version_info_lib_clear_option_key
and version_info_app_clear_option_key =
  Co.version_info_app_clear_option_key
and version_info_lib_clear_option_key =
  Co.version_info_lib_clear_option_key
and short_version_info_app_clear_option_key =
  Co.short_version_info_app_clear_option_key
and short_version_info_lib_clear_option_key =
  Co.short_version_info_lib_clear_option_key
;;

let phase_ulib_versioning = Co.phase_ulib_versioning
and phase_long_version_info = Co.phase_long_version_info
and phase_version_info = Co.phase_version_info
and phase_short_version_info = Co.phase_short_version_info
and phase_all_versioning = Co.phase_all_versioning
;;

(* Shortcuts. *)

let lift_pf pf pic kps_opt =
  pf { Co.name = pic.project_name; option_key_prefixes_opt = kps_opt; }
;;

let application_printing_facilities =
  lift_pf D.application_printing_facilities
and library_printing_facilities =
  lift_pf D.library_printing_facilities

and application_printing_facilities_with_static_tag =
  lift_pf D.application_printing_facilities_with_static_tag
and library_printing_facilities_with_static_tag =
  lift_pf D.library_printing_facilities_with_static_tag

and application_printing_facilities_with_dynamic_tags pic =
  lift_pf D.application_printing_facilities_with_dynamic_tags pic
and library_printing_facilities_with_dynamic_tags pic =
  lift_pf D.library_printing_facilities_with_dynamic_tags pic
;;
